import pytest

from spidermon.python import Interpreter
from spidermon.exceptions import InvalidExpression
from spidermon.data import Data

from .fixtures.expressions import *
from .fixtures.stats import STATS_TO_EVALUATE


@pytest.fixture
def interpreter():
    return Interpreter()


def test_syntax_errors(interpreter):
    for expression in SYNTAXERROR_EXPRESSIONS:
        with pytest.raises(SyntaxError):
            interpreter.check(expression)


def test_invalid_expressions(interpreter):
    for expression in INVALID_EXPRESSIONS:
        with pytest.raises(InvalidExpression):
            interpreter.check(expression)


def test_valid_expressions(interpreter):
    for expression in VALID_EXPRESSIONS:
        interpreter.check(expression)


def test_evaluated_expressions(interpreter):
    data = Data({"stats": Data(STATS_TO_EVALUATE)})
    for expression, result in EXPRESSIONS_TO_EVALUATE:
        assert result == interpreter.eval(
            expression, data
        ), f'Expression fails: "{expression}" != {result}'



from spidermon.exceptions import NotConfigured


class StatsMonitorMixin:
    @property
    def stats(self):
        if not self.data.stats:
            raise NotConfigured("Stats not available!")
        return self.data.stats



from spidermon import settings


class ItemResult:
    def __init__(self, item):
        self.item = item
        self.status = settings.UNDEFINED_STATUS
        self.error = None
        self.reason = None
        self.id = id(self)


class MonitorResult(ItemResult):
    name = "monitor"

    def __init__(self, item):
        super().__init__(item)
        self.status = settings.MONITOR.STATUSES.DEFAULT

    @property
    def monitor(self):
        return self.item


class ActionResult(ItemResult):
    name = "action"

    def __init__(self, item):
        super().__init__(item)
        self.status = settings.ACTION.STATUSES.DEFAULT

    @property
    def action(self):
        return self.item



class DecoratorWithAttributes:
    name = None
    attributes = {}

    def __init__(self):
        if not self.name:
            raise AttributeError("No name defined!")
        if not self.attributes:
            raise AttributeError("No attributes defined!")

    def __getattr__(self, name):
        if name not in self.attributes:
            raise AttributeError(
                "Invalid {attribute} '{name}', allowed values: {values}".format(
                    attribute=self.name,
                    name=name,
                    values=", ".join(
                        ["'%s'" % attr for attr in self.attributes.keys()]
                    ),
                )
            )
        else:
            return self.attributes[name]


class OptionsDecorator:
    @classmethod
    def set_value(cls, options_class, value_name):
        def value_decorator(value):
            def decorator(fn):
                options_class.add_or_create(fn)
                setattr(fn.options, value_name, value)
                return fn

            return decorator

        return value_decorator

    @classmethod
    def set_fixed_value(cls, options_class, value_name, value):
        def decorator(fn):
            options_class.add_or_create(fn)
            setattr(fn.options, value_name, value)
            return fn

        return decorator



import re


class StatsAnalyzer:
    def __init__(self, stats, prefix=None):
        self.stats = stats
        self.prefix = prefix or ""

    def search(self, pattern, include_matches=False):
        pattern = re.compile(self._get_pattern(pattern))
        results = {}
        for key, count in self.stats.items():
            match = pattern.match(key)
            if match:
                if include_matches:
                    results[key] = (
                        count,
                        match.group(1) if len(match.groups()) else "",
                    )
                else:
                    results[key] = count
        return results

    def _get_pattern(self, pattern):
        if self.prefix:
            return "/".join([self.prefix, pattern])
        else:
            return pattern



import logging

from spidermon.results.monitor import (
    MonitorResult,
    actions_step_required,
    monitors_step_required,
)
from spidermon.runners import MonitorRunner
from spidermon.utils.text import Message, line, line_title

LOG_MESSAGE_HEADER = "Spidermon"


class SpiderMonitorResult(MonitorResult):
    def __init__(self, spider):
        super().__init__()
        self.spider = spider

    def next_step(self):
        super().next_step()
        self.write_title()

    def finish_step(self):
        super().finish_step()
        self.log_info(line())
        if not self.step.successful:
            self.write_errors()
        self.write_run_footer()
        self.write_step_summary()

    @monitors_step_required
    def addSuccess(self, test):
        super().addSuccess(test)
        self.write_item_result(test)

    @monitors_step_required
    def addError(self, test, error):
        super().addError(test, error)
        self.write_item_result(test)

    @monitors_step_required
    def addFailure(self, test, error):
        super().addFailure(test, error)
        self.write_item_result(test)

    @monitors_step_required
    def addSkip(self, test, reason):
        super().addSkip(test, reason)
        self.write_item_result(test, reason)

    @monitors_step_required
    def addExpectedFailure(self, test, error):
        super().addExpectedFailure(test, error)
        self.write_item_result(test)

    @monitors_step_required
    def addUnexpectedSuccess(self, test):
        super().addUnexpectedSuccess(test)
        self.write_item_result(test)

    @actions_step_required
    def add_action_success(self, action):
        super().add_action_success(action)
        self.write_item_result(action)

    @actions_step_required
    def add_action_skip(self, action, reason):
        super().add_action_skip(action, reason)
        self.write_item_result(action, reason)

    @actions_step_required
    def add_action_error(self, action, error):
        super().add_action_error(action, error)
        self.write_item_result(action)

    def write_title(self):
        self.log_info(line_title(self.step.name))

    def write_item_result(self, item, extra=None):
        self.log_info(
            "%s... %s%s"
            % (item.name, self.step[item].status, " (%s)" % extra if extra else "")
        )

    def write_run_footer(self):
        self.log_info(
            "{count:d} {item_name}{plural_suffix} in {time:.3f}s".format(
                count=self.step.number_of_items,
                item_name=self.step.item_result_class.name,
                plural_suffix="" if self.step.number_of_items == 1 else "s",
                time=self.step.time_taken,
            )
        )

    def write_step_summary(self):
        summary = "OK" if self.step.successful else "FAILED"
        infos = self.step.get_infos()
        if infos and sum(infos.values()):
            summary += " (%s)" % ", ".join([f"{k}={v}" for k, v in infos.items() if v])
        self.log_info(summary)

    def write_errors(self):
        for status in self.step.error_statuses:
            for item in self.step.items_for_status(status):
                msg = Message()
                msg.write_line()
                msg.write_bold_separator()
                msg.write_line(f"{item.status}: {item.item.name}")
                msg.write_light_separator()
                msg.write(item.error)
                self.log_error(msg)

    def log_error(self, msg):
        self.log(msg, level=logging.ERROR)

    def log_info(self, msg):
        self.log(msg, level=logging.INFO)

    def log(self, msg, level=logging.DEBUG):
        self.spider.log(f"[{LOG_MESSAGE_HEADER}] {msg}", level=level)


class SpiderMonitorRunner(MonitorRunner):
    def __init__(self, spider):
        super().__init__()
        self.spider = spider

    def create_result(self):
        return SpiderMonitorResult(self.spider)



import abc
import traceback

from spidermon.exceptions import SkipAction

from .options import ActionOptionsMetaclass


class Action(metaclass=ActionOptionsMetaclass):
    """
    Base class for actions.
    """

    def __init__(self):
        self.result = None
        self.data = None

    @classmethod
    def from_crawler(cls, crawler):
        return cls(**cls.from_crawler_kwargs(crawler))

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {}

    @property
    def name(self):
        return self.options.name or self.__class__.__name__

    def run(self, result, data):
        self.result = result
        self.data = data
        result.start_action(self)
        try:
            self.run_action()
        except SkipAction as e:
            result.add_action_skip(self, e.args[0])
        except:
            result.add_action_error(self, traceback.format_exc())
        else:
            result.add_action_success(self)
        data.meta.update(self.get_meta())

    @abc.abstractmethod
    def run_action(self):
        raise NotImplementedError

    @property
    def monitors_passed(self):
        return len(self.result.monitors_passed_results) > 0

    @property
    def monitors_failed(self):
        return len(self.result.monitors_failed_results) > 0

    def get_meta(self):
        return {}

    def __repr__(self):
        return "<ACTION:({}) at {}>".format(self.name, hex(id(self)))

    def __str__(self):
        return repr(self)


class DummyAction(Action):
    def run_action(self):
        pass



import abc
from spidermon import settings


class OptionsMetaclassBase(abc.ABCMeta):
    __options_class__ = None

    def __new__(mcs, name, bases, attrs):
        cls = super().__new__(mcs, name, bases, attrs)
        if not cls.__options_class__:
            raise TypeError(
                "Options class not defined! "
                "are you trying to use OptionsMetaclassBase?"
            )
        cls.options = cls.__options_class__()
        return cls


class OptionsBase:
    __options_name__ = "options"

    @classmethod
    def add_or_create(cls, target):
        if not hasattr(target, cls.__options_name__):
            setattr(target, cls.__options_name__, cls())
            return True
        return False

    def _get_attributes(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}

    def __str__(self):
        return "<{name}:({attributes})>".format(
            name=self.__class__.__name__,
            attributes=", ".join(
                "{}={}".format(attr, getattr(self, attr))
                for attr in self._get_attributes()
            ),
        )


class MonitorOptions(OptionsBase):
    def __init__(self):
        self.name = settings.MONITOR.DEFAULT_NAME
        self.description = settings.MONITOR.DEFAULT_DESCRIPTION
        self.level = None
        self.meta = {}
        self.order = settings.MONITOR.DEFAULT_ORDER


class MonitorOptionsMetaclass(OptionsMetaclassBase):
    __options_class__ = MonitorOptions


class ActionOptions(OptionsBase):
    def __init__(self):
        self.name = None
        self.description = settings.ACTION.DEFAULT_DESCRIPTION


class ActionOptionsMetaclass(OptionsMetaclassBase):
    __options_class__ = ActionOptions



import pytest

from spidermon import Monitor


def pytest_report_header(config):
    return "Spidermon monitor filtering"


@pytest.mark.trylast
def pytest_collection_modifyitems(session, config, items):
    items[:] = [
        item
        for item in items
        if hasattr(item, "cls") and not (item.cls and issubclass(item.cls, Monitor))
    ]



import json
import logging
import requests

from spidermon.contrib.actions.templates import ActionWithTemplates
from spidermon.exceptions import NotConfigured

logger = logging.getLogger(__name__)


class SimplyTelegramClient:
    send_message_api = "https://api.telegram.org/bot{token}/sendMessage?chat_id={chat_id}&parse_mode=Markdown&text={text}"

    def __init__(self, token):
        self.token = token

    def send_message(self, message, recipient):
        api_url = self.send_message_api.format(
            token=self.token, chat_id=recipient, text=message
        )
        r = requests.get(api_url).json()
        if r.get("ok") is False:
            logger.error(
                "Failed to send message. Telegram api error: %s", json.dumps(r)
            )


class TelegramMessageManager:
    sender_token = None

    def __init__(self, sender_token=None, fake=False):
        sender_token = sender_token or self.sender_token
        if not sender_token:
            raise NotConfigured("You must provide a telegram bot token.")

        self.fake = fake
        self._client = SimplyTelegramClient(sender_token)

    def send_message(self, to, text):
        if self.fake:
            logger.info(text)
            return
        for recipient in to:
            self._client.send_message(text, recipient)


class SendTelegramMessage(ActionWithTemplates):
    message = None
    message_template = "telegram/default/message.jinja"
    recipients = None
    sender_token = None
    fake = False

    def __init__(
        self,
        sender_token=None,
        recipients=None,
        message=None,
        message_template=None,
        fake=None,
    ):
        super().__init__()

        self.fake = fake or self.fake
        self.manager = TelegramMessageManager(
            sender_token=sender_token or self.sender_token, fake=self.fake
        )

        self.recipients = recipients or self.recipients
        self.message = message or self.message
        self.message_template = message_template or self.message_template
        if not self.recipients:
            raise NotConfigured(
                "You must provide at least one recipient for the message."
            )

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {
            "sender_token": crawler.settings.get("SPIDERMON_TELEGRAM_SENDER_TOKEN"),
            "recipients": crawler.settings.getlist("SPIDERMON_TELEGRAM_RECIPIENTS"),
            "message": crawler.settings.get("SPIDERMON_TELEGRAM_MESSAGE"),
            "message_template": crawler.settings.get(
                "SPIDERMON_TELEGRAM_MESSAGE_TEMPLATE"
            ),
            "fake": crawler.settings.getbool("SPIDERMON_TELEGRAM_FAKE"),
        }

    def run_action(self):
        self.manager.send_message(to=self.recipients, text=self.get_message())

    def get_message(self):
        if self.message:
            return self.render_text_template(self.message)
        else:
            return self.render_template(self.message_template)



from spidermon.utils.field_coverage import calculate_field_coverage


def test_calculate_field_coverage_from_stats():
    spider_stats = {
        "finish_reason": "finished",
        "spidermon_item_scraped_count": 100,
        "spidermon_item_scraped_count/dict": 100,
        "spidermon_item_scraped_count/dict/author": 100,
        "spidermon_item_scraped_count/dict/author/author_url": 64,
        "spidermon_item_scraped_count/dict/author/name": 100,
        "spidermon_item_scraped_count/dict/quote": 50,
        "spidermon_item_scraped_count/dict/tags": 100,
    }

    expected_coverage = {
        "spidermon_field_coverage/dict/author": 1.0,
        "spidermon_field_coverage/dict/author/author_url": 0.64,
        "spidermon_field_coverage/dict/author/name": 1.0,
        "spidermon_field_coverage/dict/quote": 0.5,
        "spidermon_field_coverage/dict/tags": 1.0,
    }

    coverage = calculate_field_coverage(spider_stats)

    assert coverage == expected_coverage
